import inspect

from plotly import express as px

MAX_NUM_TRACES = 3

NONE_OPTION = {"label": "", "value": None}

QUALITATIVE_COLOR_SCALES = {
    name: body
    for name, body in inspect.getmembers(px.colors.qualitative)
    if isinstance(body, list)
    and not name.startswith("__")
    and not name.endswith("_r")
    and body[0].startswith("#")
}

NAMED_COLORS = [
    "aliceblue",
    "antiquewhite",
    "aqua",
    "aquamarine",
    "azure",
    "beige",
    "bisque",
    "black",
    "blanchedalmond",
    "blue",
    "blueviolet",
    "brown",
    "burlywood",
    "cadetblue",
    "chartreuse",
    "chocolate",
    "coral",
    "cornflowerblue",
    "cornsilk",
    "crimson",
    "cyan",
    "darkblue",
    "darkcyan",
    "darkgoldenrod",
    "darkgray",
    "darkgrey",
    "darkgreen",
    "darkkhaki",
    "darkmagenta",
    "darkolivegreen",
    "darkorange",
    "darkorchid",
    "darkred",
    "darksalmon",
    "darkseagreen",
    "darkslateblue",
    "darkslategray",
    "darkslategrey",
    "darkturquoise",
    "darkviolet",
    "deeppink",
    "deepskyblue",
    "dimgray",
    "dimgrey",
    "dodgerblue",
    "firebrick",
    "floralwhite",
    "forestgreen",
    "fuchsia",
    "gainsboro",
    "ghostwhite",
    "gold",
    "goldenrod",
    "gray",
    "grey",
    "green",
    "greenyellow",
    "honeydew",
    "hotpink",
    "indianred",
    "indigo",
    "ivory",
    "khaki",
    "lavender",
    "lavenderblush",
    "lawngreen",
    "lemonchiffon",
    "lightblue",
    "lightcoral",
    "lightcyan",
    "lightgoldenrodyellow",
    "lightgray",
    "lightgrey",
    "lightgreen",
    "lightpink",
    "lightsalmon",
    "lightseagreen",
    "lightskyblue",
    "lightslategray",
    "lightslategrey",
    "lightsteelblue",
    "lightyellow",
    "lime",
    "limegreen",
    "linen",
    "magenta",
    "maroon",
    "mediumaquamarine",
    "mediumblue",
    "mediumorchid",
    "mediumpurple",
    "mediumseagreen",
    "mediumslateblue",
    "mediumspringgreen",
    "mediumturquoise",
    "mediumvioletred",
    "midnightblue",
    "mintcream",
    "mistyrose",
    "moccasin",
    "navajowhite",
    "navy",
    "oldlace",
    "olive",
    "olivedrab",
    "orange",
    "orangered",
    "orchid",
    "palegoldenrod",
    "palegreen",
    "paleturquoise",
    "palevioletred",
    "papayawhip",
    "peachpuff",
    "peru",
    "pink",
    "plum",
    "powderblue",
    "purple",
    "red",
    "rosybrown",
    "royalblue",
    "saddlebrown",
    "salmon",
    "sandybrown",
    "seagreen",
    "seashell",
    "sienna",
    "silver",
    "skyblue",
    "slateblue",
    "slategray",
    "slategrey",
    "snow",
    "springgreen",
    "steelblue",
    "tan",
    "teal",
    "thistle",
    "tomato",
    "turquoise",
    "violet",
    "wheat",
    "white",
    "whitesmoke",
    "yellow",
    "yellowgreen",
]

MARKER_SYMBOLS = [
    "arrow-bar-up",
    "circle",
    "hexagram",
    "star-diamond",
    "y-up",
    "diamond-wide",
    "octagon",
    "arrow-bar-down",
    "triangle-se",
    "cross",
    "triangle-left",
    "x-thin",
    "y-right",
    "star-triangle-down",
    "asterisk",
    "triangle-nw",
    "x",
    "diamond-cross",
    "arrow-down",
    "arrow-bar-right",
    "triangle-down",
    "arrow-bar-left",
    "circle-cross",
    "square-cross",
    "triangle-right",
    "diamond-x",
    "line-ns",
    "y-left",
    "line-nw",
    "hourglass",
    "arrow-right",
    "arrow-left",
    "diamond-tall",
    "triangle-up",
    "star",
    "y-down",
    "star-triangle-up",
    "hexagon2",
    "square-x",
    "hexagon",
    "pentagon",
    "hash",
    "triangle-ne",
    "line-ew",
    "line-ne",
    "diamond",
    "star-square",
    "cross-thin",
    "circle-x",
    "bowtie",
    "triangle-sw",
    "arrow-up",
    "square",
]

SUPPORTED_GRAPH_TYPES = (
    "bar",
    "line",
    "area",
    "violin",
    "timeline",
    "pie",
    "density_heatmap",
    "scatter_matrix",
    "strip",
    "histogram",
    "density_contour",
    "box",
    "scatter",
    "scatter_ternary",
    "parallel_coordinates",
    "scatter_3d",
    "line_3d",
)

DEFAULT_GRAPH_TYPE = ""

UNSUPPORTED_PARAMS = {
    "animation_group",
    "category_orders",
    "facet_row_spacing",
    "data_frame",
    "facet_col_wrap",
    "labels",
    "range_y",
    "custom_data",
    "color_discrete_map",
    "facet_col_spacing",
    "orientation",
    "template",
    "hover_name",
    "range_x",
    "animation_frame",
    "trendline_scope",
    "symbol_map",
    "line_dash_map",
    "symbol_sequence",
    "trendline_color_override",
    "line_dash_sequence",
    "trendline",
    "color_continuous_midpoint",
    "range_color",
    "opacity",
    "render_mode",
    "pattern_shape_sequence",
    "pattern_shape_map",
    "hole",
    "size_max",
    "trendline_options",
    "range_z",
}



import random
from typing import Any

import dash_bootstrap_components as dbc
import plotly.graph_objs as go
from dash import callback_context, dcc, html
from dash.dependencies import Input, Output
from jupyter_dash import JupyterDash
from plotly.subplots import make_subplots

from kindergarten.constants import MAX_NUM_TRACES
from kindergarten.tab import Tab


class Kindergarten:
    def __init__(self, num_traces=MAX_NUM_TRACES):
        self.tabs = [Tab(tab_id=i) for i in range(num_traces)]

        self.app = JupyterDash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        self._initialize_app()

    def _initialize_app(self):
        self.app.config.suppress_callback_exceptions = True

        self.app.layout = dbc.Container(
            [
                dbc.Tabs(
                    [
                        dbc.Tab(
                            self.tabs[i].component(), label=f"Trace {i}", id=f"tab-{i}"
                        )
                        for i in range(len(self.tabs))
                    ],
                    id="tabs",
                ),
                dbc.Row(dbc.Col(dcc.Graph(id="graph"))),
                dbc.Row(
                    dbc.Col(
                        [
                            html.Div(
                                dbc.Button(
                                    "Print Code",
                                    id="print-code",
                                    color="secondary",
                                    n_clicks=0,
                                    style={"margin-top": 15},
                                )
                            ),
                            html.Div(
                                id="hidden-div",
                                style={"display": "none"},
                            ),
                        ]
                    )
                ),
            ],
            fluid=True,
            className="dash-bootstrap",
        )

        for i in range(len(self.tabs)):

            @self.app.callback(
                Output(f"selector-{i}", "children"),
                [Input(f"graph-type-{i}", "value"), Input(f"dataframe-{i}", "value")],
                prevent_initial_call=True,
            )
            def _on_graph_type_or_dataframe_change_update_selector(
                graph_type, dataframe_name
            ):
                triggered_component_id = callback_context.triggered_id
                kw, tab_id = triggered_component_id.rsplit("-", 1)

                self.tabs[int(tab_id)].update_option(
                    kw, graph_type if kw == "graph-type" else dataframe_name
                )
                return self.tabs[int(tab_id)].options_component()

        all_options = sum(
            [list(tab.options.values()) for tab in self.tabs],
            start=[],
        )
        inputs = (
            [Input(option.id, "value") for option in all_options]
            + [Input(f"graph-type-{i}", "value") for i in range(len(self.tabs))]
            + [Input(f"dataframe-{i}", "value") for i in range(len(self.tabs))]
        )
        input_names = (
            [option.id for option in all_options]
            + [f"graph-type-{i}" for i in range(len(self.tabs))]
            + [f"dataframe-{i}" for i in range(len(self.tabs))]
        )

        @self.app.callback(
            Output("graph", "figure"), inputs, prevent_initial_call=False
        )
        def _on_change_update_graph(*args) -> Any:
            triggered_component_id = callback_context.triggered_id

            if triggered_component_id is not None:
                kwargs = dict(zip(input_names, args))
                value = kwargs[triggered_component_id]

                kw, tab_id = triggered_component_id.rsplit("-", 1)
                self.tabs[int(tab_id)].update_option(kw, value)

            return self._figure()

        @self.app.callback(
            Output("hidden-div", "children"), Input("print-code", "n_clicks")
        )
        def _on_print_code(n_clicks: int):
            if n_clicks > 0:
                s = f"""
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import plotly.express as px

fig = make_subplots()
"""

                for tab in self.tabs:
                    if tab.has_figure():
                        varname = f"trace_{tab.tab_id}"
                        s += f"""
{tab.figure_str(varname)[:-1]}
fig.add_traces(list({varname}.select_traces()))
fig.update_layout({varname}.layout)
"""

                for tab in self.tabs:
                    if tab.layout_kwargs():
                        s += f"\nfig.update_layout(**{tab.layout_kwargs()})"

                s += "\nfig.update_layout(showlegend=True)"
                s += "\nfig.show()"
                print(s)

            return html.Div()

    def _figure(self) -> go.Figure:
        fig = make_subplots()

        for tab in self.tabs:
            f = tab.figure()
            fig.add_traces(list(f.select_traces()))
            # noinspection PyTypeChecker
            fig.update_layout(f.layout)

        for tab in self.tabs:
            fig.update_layout(tab.layout_kwargs())

        return fig

    def run(self):
        return self.app.run_server(
            port=random.randint(2000, 4000),
            mode="inline",
        )


def plot(num_traces=MAX_NUM_TRACES):
    Kindergarten(num_traces).run()


__all__ = ["plot"]



import collections
import inspect
import warnings
from abc import ABC, abstractmethod
from typing import Dict, Any, Tuple, Type

import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
from dash import dcc
from dash import html
from dash.development.base_component import Component

from kindergarten.constants import (
    NONE_OPTION,
    QUALITATIVE_COLOR_SCALES,
    SUPPORTED_GRAPH_TYPES,
    UNSUPPORTED_PARAMS,
    NAMED_COLORS,
    MARKER_SYMBOLS,
)


def to_options(values) -> Any:
    return [
        {
            "label": value.replace("_", " ") if isinstance(value, str) else value,
            "value": value,
        }
        for value in sorted(values, key=lambda val: str(val))
    ]


def column_options(df, include_none=True, include_name_if_present=True):
    if include_name_if_present and (df.columns.name is not None):
        cols = to_options(list(df.columns) + [df.columns.name])
    else:
        cols = to_options(list(df.columns))

    if include_none:
        return [NONE_OPTION] + cols
    else:
        return cols


def nth_numeric_column_name(df, n):
    numeric_cols = [
        df.columns[i]
        for i in range(len(df.columns))
        if pd.api.types.is_numeric_dtype(df.dtypes[i])
    ]
    try:
        return numeric_cols[n]
    except IndexError:
        return None


class GraphOption(ABC):
    basic = False
    is_px_keyword = True
    valid_graph_types = ()
    keyword = ""
    label = ""

    def __init__(self, df: pd.DataFrame, option_id: int):
        self.df: pd.DataFrame = df
        self.id = f"{self.keyword}-{option_id}"
        self.component: Component = self._build_component()
        self.hidden_component: Component = self._build_hidden_component()

    def _build_hidden_component(self) -> Component:
        return self._build_component(style={"display": "none"})

    def _build_component(self, **kwargs) -> Component:
        return html.Div(
            [html.Label([self.label, self._build_inner_component()])], **kwargs
        )

    def kwarg(self, value) -> Dict[str, Any]:
        return {self.keyword: value}

    def default_kwarg(self) -> Dict[str, Any]:
        return {self.keyword: self.default_kwarg_value()}

    def is_valid_for_graph_type(self, graph_type: str) -> bool:
        return graph_type in self.valid_graph_types

    @abstractmethod
    def _build_inner_component(self) -> Component:
        pass

    @abstractmethod
    def default_kwarg_value(self):
        pass


def build_graph_option(
    *,
    _basic: bool,
    _keyword: str,
    _label: str,
    _build_inner_component_callable,
    _default_kwarg_value_callable,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    class C(GraphOption):
        basic = _basic
        keyword = _keyword
        label = _label
        is_px_keyword = _is_px_keyword
        valid_graph_types = _valid_graph_types

        def _build_inner_component(self) -> Component:
            return _build_inner_component_callable(self)

        def default_kwarg_value(self):
            return _default_kwarg_value_callable(self)

    return C


def build_select_graph_option(
    *,
    _keyword: str,
    _label: str,
    _basic: bool = False,
    _default_kwarg_value_callable: Any = lambda self: None,
    _select_options_callable: Any = lambda self: column_options(self.df),
    _required: bool = False,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    def _build_inner_component_callable(self: GraphOption) -> Component:
        return dbc.Select(
            id=self.id,
            required=_required,
            value=self.default_kwarg_value(),
            options=_select_options_callable(self),
        )

    return build_graph_option(
        _basic=_basic,
        _keyword=_keyword,
        _label=_label,
        _build_inner_component_callable=_build_inner_component_callable,
        _default_kwarg_value_callable=_default_kwarg_value_callable,
        _is_px_keyword=_is_px_keyword,
        _valid_graph_types=_valid_graph_types,
    )


def build_multiselect_graph_option(
    *,
    _keyword: str,
    _label: str,
    _basic: bool = False,
    _default_kwarg_value_callable=lambda self: None,
    _select_options_callable: Any = lambda self: column_options(
        self.df, include_none=False
    ),
    _required: bool = False,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    def _build_inner_component_callable(self: GraphOption) -> Component:
        return dcc.Dropdown(
            id=self.id,
            clearable=_required,
            value=self.default_kwarg_value(),
            options=_select_options_callable(self),
            multi=True,
            style={"min-width": 150},
        )

    return build_graph_option(
        _basic=_basic,
        _keyword=_keyword,
        _label=_label,
        _build_inner_component_callable=_build_inner_component_callable,
        _default_kwarg_value_callable=_default_kwarg_value_callable,
        _is_px_keyword=_is_px_keyword,
        _valid_graph_types=_valid_graph_types,
    )


def build_checklist_graph_option(
    *,
    _keyword: str,
    _label: str,
    _basic: bool = False,
    _default_kwarg_value_callable=lambda self: [],
    _checklist_options_callable=lambda self: column_options(
        self.df, include_none=False
    ),
    _required: bool = False,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    def _build_inner_component_callable(self: GraphOption) -> Component:
        return dbc.Checklist(
            id=self.id,
            value=self.default_kwarg_value(),
            options=_checklist_options_callable(self),
            switch=True,
            inline=True,
        )

    return build_graph_option(
        _basic=_basic,
        _keyword=_keyword,
        _label=_label,
        _build_inner_component_callable=_build_inner_component_callable,
        _default_kwarg_value_callable=_default_kwarg_value_callable,
        _is_px_keyword=_is_px_keyword,
        _valid_graph_types=_valid_graph_types,
    )


def build_switch_graph_option(
    *,
    _keyword: str,
    _label: str,
    _basic: bool = False,
    _default_kwarg_value_callable=lambda self: False,
    _required: bool = False,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    def _build_inner_component_callable(self: GraphOption) -> Component:
        return dbc.Switch(
            id=self.id,
            value=self.default_kwarg_value(),
            style={"margin-left": 17},
        )

    return build_graph_option(
        _basic=_basic,
        _keyword=_keyword,
        _label=_label,
        _build_inner_component_callable=_build_inner_component_callable,
        _default_kwarg_value_callable=_default_kwarg_value_callable,
        _is_px_keyword=_is_px_keyword,
        _valid_graph_types=_valid_graph_types,
    )


def build_numeric_graph_option(
    *,
    _keyword: str,
    _label: str,
    _min: float,
    _max: float,
    _step: float,
    _basic: bool = False,
    _default_kwarg_value_callable=lambda self: None,
    _required: bool = False,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    def _build_inner_component_callable(self: GraphOption) -> Component:
        return dbc.Input(
            type="number",
            min=_min,
            max=_max,
            step=_step,
            id=self.id,
            value=self.default_kwarg_value(),
            debounce=True,
        )

    return build_graph_option(
        _basic=_basic,
        _keyword=_keyword,
        _label=_label,
        _build_inner_component_callable=_build_inner_component_callable,
        _default_kwarg_value_callable=_default_kwarg_value_callable,
        _is_px_keyword=_is_px_keyword,
        _valid_graph_types=_valid_graph_types,
    )


def build_text_graph_option(
    *,
    _keyword: str,
    _label: str,
    _basic: bool = False,
    _default_kwarg_value_callable: Any = lambda self: "",
    _required: bool = False,
    _is_px_keyword: bool = True,
    _valid_graph_types: Tuple[str, ...] = (),
):
    def _build_inner_component_callable(self: GraphOption) -> Component:
        return dbc.Input(
            type="text", id=self.id, value=self.default_kwarg_value(), debounce=True
        )

    return build_graph_option(
        _basic=_basic,
        _keyword=_keyword,
        _label=_label,
        _build_inner_component_callable=_build_inner_component_callable,
        _default_kwarg_value_callable=_default_kwarg_value_callable,
        _is_px_keyword=_is_px_keyword,
        _valid_graph_types=_valid_graph_types,
    )


XAxis = build_select_graph_option(
    _keyword="x",
    _label="X-Axis",
    _basic=True,
)

Dimensions = build_multiselect_graph_option(
    _keyword="dimensions",
    _label="Dimensions",
    _basic=True,
)

XStart = build_select_graph_option(
    _keyword="x_start",
    _label="X-Axis Start Values",
    _basic=True,
)

XEnd = build_select_graph_option(
    _keyword="x_end",
    _label="X-Axis End Values",
    _basic=True,
)

YAxis = build_multiselect_graph_option(
    _keyword="y",
    _label="Y-Axis",
    _basic=True,
)

ZAxis = build_select_graph_option(
    _keyword="z",
    _label="Z-Axis",
    _basic=True,
)

Title = build_text_graph_option(
    _keyword="title",
    _label="Title",
)

A = build_select_graph_option(
    _keyword="a",
    _label="A",
    _basic=True,
)

B = build_select_graph_option(
    _keyword="b",
    _label="B",
    _basic=True,
)

C = build_select_graph_option(
    _keyword="c",
    _label="C",
    _basic=True,
)

Names = build_select_graph_option(_keyword="names", _label="Names", _basic=True)

Values = build_select_graph_option(_keyword="values", _label="Values", _basic=True)

LineGroup = build_select_graph_option(_keyword="line_group", _label="Group By")

LineDash = build_select_graph_option(_keyword="line_dash", _label="Dash Type")

Color = build_select_graph_option(_keyword="color", _label="Color")

LineColor = build_select_graph_option(
    _keyword="line_color",
    _label="Line Color",
    _select_options_callable=lambda self: [NONE_OPTION] + to_options(NAMED_COLORS),
    _is_px_keyword=False,
    _valid_graph_types=("line",),
)

PatternShape = build_select_graph_option(
    _keyword="pattern_shape", _label="Pattern Shape"
)

# noinspection PyTypeChecker
BarMode = build_select_graph_option(
    _keyword="barmode",
    _label="Bar Mode",
    _default_kwarg_value_callable=lambda self: "relative",
    _select_options_callable=lambda self: to_options(("relative", "group", "overlay")),
)

# noinspection PyTypeChecker
BoxMode = build_select_graph_option(
    _keyword="boxmode",
    _label="Box Mode",
    _default_kwarg_value_callable=lambda self: "group",
    _select_options_callable=lambda self: to_options(("group", "overlay")),
)

# noinspection PyTypeChecker
ViolinMode = build_select_graph_option(
    _keyword="violinmode",
    _label="Violin Mode",
    _default_kwarg_value_callable=lambda self: "group",
    _select_options_callable=lambda self: to_options(("group", "overlay")),
)

# noinspection PyTypeChecker
StripMode = build_select_graph_option(
    _keyword="stripmode",
    _label="Strip Mode",
    _default_kwarg_value_callable=lambda self: "group",
    _select_options_callable=lambda self: to_options(("group", "overlay")),
)

# noinspection PyTypeChecker
Points = build_select_graph_option(
    _keyword="points",
    _label="Points to Show",
    _default_kwarg_value_callable=lambda self: "outliers",
    _select_options_callable=lambda self: to_options(
        ("outliers", "suspectedoutliers", "all", False)
    ),
)

ColorContinuousScale = build_select_graph_option(
    _keyword="color_continuous_scale",
    _label="Color Scale",
    _select_options_callable=lambda self: to_options(px.colors.named_colorscales()),
)

# noinspection PyTypeChecker
ColorDiscreteSequence = build_select_graph_option(
    _keyword="color_discrete_sequence",
    _label="Color Sequence",
    _default_kwarg_value_callable=lambda self: QUALITATIVE_COLOR_SCALES["Plotly"],
    _select_options_callable=lambda self: [
        {"label": name, "value": body}
        for name, body in QUALITATIVE_COLOR_SCALES.items()
    ],
)

Size = build_select_graph_option(_keyword="size", _label="Size")

Symbol = build_select_graph_option(_keyword="symbol", _label="Symbol")

XError = build_select_graph_option(_keyword="error_x", _label="X-Axis Error Bars")

YError = build_select_graph_option(_keyword="error_y", _label="Y-Axis Error Bars")

ZError = build_select_graph_option(_keyword="error_z", _label="Z-Axis Error Bars")

XErrorMinus = build_select_graph_option(
    _keyword="error_x_minus", _label="X-Axis Error Bars in Negative Direction"
)

YErrorMinus = build_select_graph_option(
    _keyword="error_y_minus", _label="Y-Axis Error Bars in Negative Direction"
)

ZErrorMinus = build_select_graph_option(
    _keyword="error_z_minus", _label="Z-Axis Error Bars in Negative Direction"
)

HoverData = build_checklist_graph_option(_keyword="hover_data", _label="Show on Hover")

Marginal = build_select_graph_option(
    _keyword="marginal",
    _label="Marginal Distribution",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("histogram", "rug", "box", "violin")),
)

MarginalX = build_select_graph_option(
    _keyword="marginal_x",
    _label="X-Axis Marginal Distribution",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("histogram", "rug", "box", "violin")),
)

MarginalY = build_select_graph_option(
    _keyword="marginal_y",
    _label="Y-Axis Marginal Distribution",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("histogram", "rug", "box", "violin")),
)

FacetCol = build_select_graph_option(_keyword="facet_col", _label="Facet Column")

FacetRow = build_select_graph_option(_keyword="facet_row", _label="Facet Row")

LogX = build_switch_graph_option(_keyword="log_x", _label="Logarithmic X-Axis")

LogY = build_switch_graph_option(_keyword="log_y", _label="Logarithmic Y-Axis")

LogZ = build_switch_graph_option(_keyword="log_z", _label="Logarithmic Z-Axis")

Markers = build_switch_graph_option(_keyword="markers", _label="Markers")

MarkerColor = build_select_graph_option(
    _keyword="marker_color",
    _label="Marker Color",
    _select_options_callable=lambda self: [NONE_OPTION] + to_options(NAMED_COLORS),
    _is_px_keyword=False,
    _valid_graph_types=("scatter",),
)

MarkerSymbol = build_select_graph_option(
    _keyword="marker_symbol",
    _label="Marker Symbol",
    _select_options_callable=lambda self: [NONE_OPTION] + to_options(MARKER_SYMBOLS),
    _is_px_keyword=False,
    _valid_graph_types=("scatter", "line"),
)

LegendName = build_text_graph_option(
    _keyword="name",
    _label="Legend Name",
    _is_px_keyword=False,
    _valid_graph_types=("scatter", "line"),
    _default_kwarg_value_callable=lambda self: None,
)

Notched = build_switch_graph_option(_keyword="notched", _label="Notched Boxes")

Cumulative = build_switch_graph_option(_keyword="cumulative", _label="Cumulative")

TextAuto = build_switch_graph_option(_keyword="text_auto", _label="Show Text")

Box = build_switch_graph_option(_keyword="box", _label="Show Box")

NBins = build_numeric_graph_option(
    _keyword="nbins", _label="Bins", _min=0, _max=1000, _step=1
)

NBinsX = build_numeric_graph_option(
    _keyword="nbinsx", _label="X Bins", _min=0, _max=1000, _step=1
)

NBinsY = build_numeric_graph_option(
    _keyword="nbinsy", _label="Y Bins", _min=0, _max=1000, _step=1
)

MarkerSize = build_numeric_graph_option(
    _keyword="marker_size",
    _label="Marker Size",
    _min=0,
    _max=100,
    _step=1,
    _is_px_keyword=False,
    _valid_graph_types=("scatter", "line"),
)

LineShape = build_select_graph_option(
    _keyword="line_shape",
    _label="Line Shape",
    _select_options_callable=lambda self: to_options(("linear", "spline")),
)

GroupNorm = build_select_graph_option(
    _keyword="groupnorm",
    _label="Normalization",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("fraction", "percent")),
)

BarNorm = build_select_graph_option(
    _keyword="barnorm",
    _label="Bar Normalization",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("fraction", "percent")),
)

HistNorm = build_select_graph_option(
    _keyword="histnorm",
    _label="Histogram Normalization",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("percent", "probability", "density", "probability density")),
)

# noinspection PyTypeChecker
HistFunc = build_select_graph_option(
    _keyword="histfunc",
    _label="Aggregation Function",
    _default_kwarg_value_callable=lambda self: "count",
    _select_options_callable=lambda self: [NONE_OPTION]
    + to_options(("count", "sum", "avg", "min", "max")),
)

Text = build_select_graph_option(_keyword="text", _label="Text")

Base = build_select_graph_option(_keyword="base", _label="Base Position")

Width = build_numeric_graph_option(
    _keyword="width", _label="Width", _min=0, _max=10_000, _step=50
)

Height = build_numeric_graph_option(
    _keyword="height", _label="Height", _min=0, _max=10_000, _step=50
)

XAxisTitle = build_text_graph_option(
    _keyword="xaxis_title",
    _label="X-Axis Title",
    _valid_graph_types=SUPPORTED_GRAPH_TYPES,
    _is_px_keyword=False,
)

YAxisTitle = build_text_graph_option(
    _keyword="yaxis_title",
    _label="Y-Axis Title",
    _valid_graph_types=SUPPORTED_GRAPH_TYPES,
    _is_px_keyword=False,
)

ZAxisTitle = build_text_graph_option(
    _keyword="zaxis_title",
    _label="Z-Axis Title",
    _valid_graph_types=("scatter_3d", "line_3d"),
    _is_px_keyword=False,
)

LegendTitle = build_text_graph_option(
    _keyword="legend_title",
    _label="Legend Title",
    _valid_graph_types=SUPPORTED_GRAPH_TYPES,
    _is_px_keyword=False,
)

TitleFontSize = build_numeric_graph_option(
    _keyword="title_font_size",
    _label="Title Font Size",
    _valid_graph_types=SUPPORTED_GRAPH_TYPES,
    _is_px_keyword=False,
    _min=0,
    _max=100,
    _step=1,
)

GRAPH_OPTIONS: Tuple[Type["GraphOption"], ...] = tuple(GraphOption.__subclasses__())

param_to_graph_types = collections.defaultdict(set)

for graph_type in SUPPORTED_GRAPH_TYPES:
    signature = inspect.signature(getattr(px, graph_type))
    for param in signature.parameters.keys():
        if param not in UNSUPPORTED_PARAMS:
            param_to_graph_types[param].add(graph_type)

params_without_implementation = set(param_to_graph_types.keys()) - set(
    graph_option.keyword for graph_option in GRAPH_OPTIONS
)
if params_without_implementation:
    warnings.warn(
        f"The following parameters "
        f"are not ignored but do not "
        f"have an implementation: "
        f"{params_without_implementation}"
    )

for graph_option in GRAPH_OPTIONS:
    graph_option.valid_graph_types += tuple(param_to_graph_types[graph_option.keyword])

PX_KEYWORDS = {option.keyword for option in GRAPH_OPTIONS if option.is_px_keyword}
LAYOUT_KEYWORDS = {"xaxis_title", "yaxis_title", "legend_title", "title_font_size"}
TRACES_KEYWORDS = (
    {option.keyword for option in GRAPH_OPTIONS} - PX_KEYWORDS - LAYOUT_KEYWORDS
)



from typing import Dict, Any, List, Tuple

import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from dash import html
from dash.development.base_component import Component

from kindergarten.constants import NONE_OPTION, SUPPORTED_GRAPH_TYPES, DEFAULT_GRAPH_TYPE
from kindergarten.graph_options import (
    GRAPH_OPTIONS,
    GraphOption,
    PX_KEYWORDS,
    LAYOUT_KEYWORDS,
    TRACES_KEYWORDS,
    to_options,
)


class Tab:
    def __init__(self, tab_id: int):
        self.tab_id = tab_id
        self.graph_kwargs: Dict[str, Any] = {}
        self.graph_type = DEFAULT_GRAPH_TYPE
        self.df_name = None
        self.options: Dict[str, GraphOption] = {
            option.keyword: option(pd.DataFrame(), self.tab_id)
            for option in GRAPH_OPTIONS
        }
        self.graph_type_component = self._build_graph_type_component()
        self.dataframe_component = self._build_dataframe_component()

        self._init_graph_kwargs()

    def update_option(self, kw: str, value: Any):
        if kw == "graph-type":
            self.update_graph_type(value)
        elif kw == "dataframe":
            self.update_dataframe(value)
        else:
            # dbc or dash turn option "value"
            # fields into strings; we recover the list here.
            if kw == "color_discrete_sequence":
                value = value.split(",")

            if kw == "y" and len(value) == 1:
                value = value[0]

            # dbc or dash turn option "value" fields
            # into strings, so {"label": None, "value": None}
            # leads to value being an empty string
            # ('') instead of None.
            if value == "" or (kw == "y" and value == []):
                value = None

            self.graph_kwargs.update(self.options[kw].kwarg(value))

    def update_graph_type(self, graph_type: str):
        self.graph_type = graph_type
        self._reset_graph_kwargs()

    def update_dataframe(self, df_name: str):
        import __main__

        self.df_name = df_name
        self.options: Dict[str, GraphOption] = {
            option.keyword: option(getattr(__main__, self.df_name), self.tab_id)
            for option in GRAPH_OPTIONS
        }

    def figure(self) -> go.Figure:
        if not self.has_figure():
            return go.Figure()

        px_kwargs, update_traces_kwargs, _ = self._figure_kwargs()

        import __main__

        fig = getattr(px, self.graph_type)(getattr(__main__, self.df_name), **px_kwargs)
        fig.update_traces(**update_traces_kwargs)

        if self.graph_type in ("scatter", "line"):
            try:
                fig.update_traces(textposition="bottom right")
            except Exception:
                pass

        # If we don't do this, Plotly doesn't show the legend for single traces
        for d in fig["data"]:
            d["showlegend"] = True

        return fig

    def layout_kwargs(self):
        _, _, update_layout_kwargs = self._figure_kwargs()
        return update_layout_kwargs

    def figure_str(self, varname: str) -> str:
        px_kwargs, update_traces_kwargs, _ = self._figure_kwargs(ignore_defaults=True)

        s = f"# Trace {self.tab_id}\n"

        if px_kwargs:
            s += f"{varname} = px.{self.graph_type}({self.df_name}, **{px_kwargs})\n"
        else:
            s += f"{varname} = px.{self.graph_type}({self.df_name})\n"
        if update_traces_kwargs:
            s += f"{varname}.update_traces(**{update_traces_kwargs})\n"

        return s

    def has_figure(self) -> bool:
        return bool(self.graph_type)

    def _figure_kwargs(self, ignore_defaults: bool = True) -> Tuple[Dict, Dict, Dict]:
        if ignore_defaults:
            graph_keywords = set(
                key
                for key, val in self.graph_kwargs.items()
                if val != self.options[key].default_kwarg_value()
            )
        else:
            graph_keywords = set(self.graph_kwargs.keys())

        layout_keywords = graph_keywords & LAYOUT_KEYWORDS
        px_keywords = graph_keywords & PX_KEYWORDS
        traces_keywords = graph_keywords & TRACES_KEYWORDS

        px_kwargs = {kw: self.graph_kwargs[kw] for kw in px_keywords}
        update_traces_kwargs = {kw: self.graph_kwargs[kw] for kw in traces_keywords}
        update_layout_kwargs = {kw: self.graph_kwargs[kw] for kw in layout_keywords}

        return px_kwargs, update_traces_kwargs, update_layout_kwargs

    def component(self):
        return html.Div(
            dbc.Card(
                [
                    dbc.Row(
                        [
                            dbc.Col(html.Div(self.dataframe_component)),
                            dbc.Col(html.Div(self.graph_type_component)),
                        ]
                    ),
                    html.Div(
                        self.options_component(),
                        id=self.add_tab_id("selector"),
                    ),
                ],
                body=True,
            ),
        )

    def add_tab_id(self, component_id: str):
        return f"{component_id}-{self.tab_id}"

    @staticmethod
    def _build_basic_component(basic_option_components: List[Component]):
        rows = []
        for i in range(0, len(basic_option_components), 3):
            cols = []
            for component in basic_option_components[i : i + 3]:
                cols.append(dbc.Col(component))

            rows.append(dbc.Row(cols))

        return html.Div(rows)

    @staticmethod
    def _build_extended_component(extended_option_components: List[Component]):
        rows = []
        for i in range(0, len(extended_option_components), 3):
            cols = []
            for component in extended_option_components[i : i + 3]:
                cols.append(dbc.Col(component))

            rows.append(dbc.Row(cols))

        return html.Div(
            dbc.Accordion(
                [
                    dbc.AccordionItem(html.Div(rows), title="More Options"),
                ],
                start_collapsed=True,
            ),
            style={"padding-top": 10},
        )

    def options_component(self):
        (
            basic_option_components,
            extended_option_components,
            hidden_option_components,
        ) = ([], [], [])

        for option in self.options.values():
            if self.graph_type not in option.valid_graph_types:
                hidden_option_components.append(option.hidden_component)
            elif option.basic:
                basic_option_components.append(option.component)
            else:
                extended_option_components.append(option.component)

        basic_component = self._build_basic_component(basic_option_components)
        extended_component = self._build_extended_component(extended_option_components)

        return [
            basic_component,
            extended_component,
        ] + hidden_option_components

    def _build_graph_type_component(self) -> html.Label:
        return html.Label(
            [
                "Graph Type",
                dbc.Select(
                    id=self.add_tab_id("graph-type"),
                    required=True,
                    value=DEFAULT_GRAPH_TYPE,
                    options=[NONE_OPTION] + to_options(SUPPORTED_GRAPH_TYPES),
                ),
            ]
        )

    def _build_dataframe_component(self) -> html.Label:
        import __main__

        dataframes = [
            var
            for var in dir(__main__)
            if isinstance(getattr(__main__, var), pd.DataFrame)
        ]

        return html.Label(
            [
                "Dataframe",
                dbc.Select(
                    id=self.add_tab_id("dataframe"),
                    required=True,
                    value=None,
                    options=[NONE_OPTION] + to_options(dataframes),
                ),
            ]
        )

    def _reset_graph_kwargs(self):
        self.graph_kwargs.clear()

        for option in self.options.values():
            if self.graph_type in option.valid_graph_types:
                self.graph_kwargs.update(option.default_kwarg())

    def _init_graph_kwargs(self):
        self._reset_graph_kwargs()



#!/usr/bin/env python

from setuptools import setup, find_packages

version='0.0.3'

requirements = [
    "dash-bootstrap-components>=1.2.0",
    "dash>=2.6.0",
    "pandas>=1.3.5",
    "plotly>=5.9.0",
    "jupyter-dash>=0.4.2",
]

test_requirements = ["pip", "bump2version", "wheel", "watchdog", "black", "pytest"]

setup(
    author="Henri Froese",
    author_email="henri.froese@yahoo.com",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Kindergarten is a UI on top of Plotly to easily visualize Pandas DataFrames.",
    install_requires=requirements,
    license="MIT license",
    long_description="",
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords="kindergarten",
    name="kindergarten",
    packages=find_packages(include=["kindergarten", "kindergarten.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/henrifroese/kindergarten",
    version=version,
    zip_safe=False,
)
